""":mod:`autotweet.daemons` --- Learning, Answering your tweets
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module learns your tweets and answering it automatically.

"""
from __future__ import unicode_literals
import re
import time
import tweepy

from .logger_factory import get_logger
from .learning import NoAnswerError, DataCollection
from .twitter import (CONSUMER_KEY, CONSUMER_SECRET, OAuthToken, get_api,
                      get_full_text, strip_tweet)


MY_CLIENT_NAME = 'learn your tweet'
IGNORE_PATTERN = re.compile(r'(@\w+\s+)*@\w+\s{2,}|^[^@]')
collector_logger = get_logger('collector')

MENTION_PATTERN = re.compile(r'(?<=\B@)\w+')
DEFAULT_THRESHOLD = 0.3
answer_logger = get_logger('answer')


def check_ignore(status):
    if hasattr(status, 'retweeted_status'):
        return True
    if status.source == MY_CLIENT_NAME:
        return True
    if IGNORE_PATTERN.match(get_full_text(status)):
        return True

    return False


class CollectorMentionListener(tweepy.streaming.StreamListener):

    def __init__(self, api, db_url):
        super(CollectorMentionListener, self).__init__()
        self.api = api
        self.data_collection = DataCollection(db_url)
        self.me = api.me()

    def on_status(self, status):
        if check_ignore(status):
            return True

        if status.user.id == self.me.id and status.in_reply_to_status_id:
            original_status = self.api.get_status(
                status.in_reply_to_status_id,
                tweet_mode='extended'
            )

            question = strip_tweet(get_full_text(original_status))
            answer = strip_tweet(get_full_text(status), remove_url=False)

            if question and answer:
                self.data_collection.add_document(question, answer)

        return True


def collector_polling_timeline(api, db_url):
    data_collection = DataCollection(db_url)
    me = api.me()
    last_id = me.status.id

    collector_logger.debug('tracking from status id: {0}'.format(last_id))
    while 1:
        time.sleep(60)
        collector_logger.debug('polling from status id: {0}'.format(last_id))
        statuses = me.timeline(since_id=last_id)
        if statuses:
            statuses.reverse()
            last_id = statuses[-1].id
        else:
            continue

        for status in statuses:
            if check_ignore(status):
                continue
            if not status.in_reply_to_status_id:
                continue

            original_status = api.get_status(
                status.in_reply_to_status_id,
                tweet_mode='extended'
            )

            question = strip_tweet(get_full_text(original_status))
            answer = strip_tweet(get_full_text(status), remove_url=False)

            if question and answer:
                data_collection.add_document(question, answer)


def import_timeline(token, db_url, count):
    api = get_api(token)

    data_collection = DataCollection(db_url)
    me = api.me()

    statuses = me.timeline(count=count)
    statuses.reverse()

    for status in statuses:
        if check_ignore(status):
            continue
        if not status.in_reply_to_status_id:
            continue

        try:
            original_status = api.get_status(
                status.in_reply_to_status_id,
                tweet_mode='extended'
            )
        except:
            continue

        question = strip_tweet(get_full_text(original_status))
        answer = strip_tweet(get_full_text(status), remove_url=False)

        if question and answer:
            data_collection.add_document(question, answer)


def import_tweet(token, db_url, tweet_url):
    api = get_api(token)
    data_collection = DataCollection(db_url)
    pattern = re.compile(r'https?://twitter.com/\w+/status/(?P<id>\d+)')

    status = api.get_status(
        pattern.search(tweet_url).group('id'),
        tweet_mode='extended'
    )
    status_reply_to = api.get_status(
        status.in_reply_to_status_id,
        tweet_mode='extended'
    )
    text = status_reply_to.full_text
    reply = status.full_text

    data_collection.add_document(
        strip_tweet(text),
        strip_tweet(reply, remove_url=False)
    )


def learning_daemon(token, db_url, streaming=False):
    if not isinstance(token, OAuthToken):
        token = OAuthToken.from_string(token)

    auth = tweepy.OAuthHandler(CONSUMER_KEY, CONSUMER_SECRET)
    auth.set_access_token(token.key, token.secret)
    api = tweepy.API(auth)

    if streaming:
        listener = CollectorMentionListener(api, db_url)

        stream = tweepy.Stream(auth, listener)
        stream.userstream()
    else:
        collector_polling_timeline(api, db_url)


# Moved from answer


def get_friends(api):
    followings = set(user.screen_name for user in api.friends())
    followers = set(user.screen_name for user in api.followers())
    friends = followings | followers
    return friends


def get_mentions(status, friends):
    user_name = status.user.screen_name
    mentions = set(MENTION_PATTERN.findall(get_full_text(status)))
    mentions.discard(user_name)
    mentions = mentions & friends
    mentions = [user_name] + list(mentions)
    mentions = map(lambda x: '@' + x, mentions)
    return mentions


class AnswerMentionListener(tweepy.streaming.StreamListener):
    threshold = DEFAULT_THRESHOLD
    friends_timeout = 60
    friends_updated = None

    def __init__(self, api, db_url, threshold=None):
        super(AnswerMentionListener, self).__init__()
        self.api = api
        self.data_collection = DataCollection(db_url)
        self.me = api.me()

        if threshold:
            self.threshold = threshold

    def get_friends(self):
        if self.friends_updated is None or \
                time.time() - self.friends_updated > self.friends_timeout:
            self.friends = get_friends(self.api)
            self.friends_updated = time.time()

        return self.friends

    def on_status(self, status):
        if hasattr(status, 'retweeted_status') or status.user.id == self.me.id:
            return True

        mentions = get_mentions(status, self.get_friends())

        question = strip_tweet(get_full_text(status))
        status_id = status.id

        try:
            answer, ratio = self.data_collection.get_best_answer(question)
        except NoAnswerError:
            return True

        if status.in_reply_to_user_id == self.me.id or ratio >= self.threshold:
            answer_logger.info('@{0.user.screen_name}: {0.text} -> {1}'.format(
                status, answer
            ))
            try:
                self.api.update_status(
                    status='{0} {1}'.format(' '.join(mentions), answer),
                    in_reply_to_status_id=status_id)
            except tweepy.error.TweepError as e:
                answer_logger.error('Failed to update status: {0}'.format(
                    e.message
                ))


def answer_polling_timeline(api, db_url, threshold=None):
    data_collection = DataCollection(db_url)
    me = api.me()
    threshold = threshold or DEFAULT_THRESHOLD

    last_id = api.home_timeline(count=1)[0].id
    answer_logger.debug('tracking from status id: {0}'.format(last_id))

    while 1:
        time.sleep(60)
        friends = get_friends(api)

        answer_logger.debug('polling from status id: {0}'.format(last_id))
        if not threshold:
            statuses = api.mentions_timeline(since_id=last_id)
        else:
            home_timeline = api.home_timeline(since_id=last_id)
            mentions_timeline = api.mentions_timeline(since_id=last_id)
            home_ids = [status.id for status in home_timeline]

            statuses = home_timeline + [status for status in mentions_timeline
                                        if status.id not in home_ids]

        statuses = filter(lambda x: not hasattr(x, 'retweeted_status') and
                          x.user.id != me.id,
                          statuses)

        if statuses:
            statuses.reverse()
            last_id = statuses[-1].id
        else:
            continue

        for status in statuses:
            question = strip_tweet(get_full_text(status))
            mentions = get_mentions(status, friends)

            try:
                (answer, ratio) = data_collection.get_best_answer(question)
            except NoAnswerError:
                pass

            if (status.in_reply_to_user_id == me.id) or \
                    (status.user.id != me.id and ratio >= threshold):
                answer_logger.info(
                    '@{0.user.screen_name}: {0.text} -> {1}'.format(
                        status, answer
                    ))
                try:
                    api.update_status(
                        status='{0} {1}'.format(' '.join(mentions), answer),
                        in_reply_to_status_id=status.id)
                except tweepy.error.TweepError as e:
                    answer_logger.error('Failed to update status: {0}'.format(
                        e.message
                    ))


def answer_daemon(token, db_url, streaming=False, threshold=None):
    if not isinstance(token, OAuthToken):
        token = OAuthToken.from_string(token)

    auth = tweepy.OAuthHandler(CONSUMER_KEY, CONSUMER_SECRET)
    auth.set_access_token(token.key, token.secret)
    api = tweepy.API(auth)

    if streaming:
        listener = AnswerMentionListener(api, db_url, threshold=threshold)

        stream = tweepy.Stream(auth, listener)
        stream.userstream()
    else:
        answer_polling_timeline(api, db_url, threshold)



from __future__ import unicode_literals
import math
import random

from .compat import PY3, to_unicode
from .database import GRAM_LENGTH, Document, Gram, get_session
from .logger_factory import get_logger

__all__ = ('NoAnswerError', 'DataCollection')

logger = get_logger('learning')


def make_string(string):
    if PY3:
        if isinstance(string, memoryview):
            string = string.tobytes()
    else:
        if isinstance(string, buffer):  # noqa
            string = str(string)

    return to_unicode(string)


class DataCollection(object):

    def __init__(self, db_url):
        self.Session = get_session(db_url)

    def add_document(self, question, answer):
        """Add question answer set to DB.

        :param question: A question to an answer
        :type question: :class:`str`

        :param answer: An answer to a question
        :type answer: :class:`str`

        """
        question = question.strip()
        answer = answer.strip()

        session = self.Session()

        if session.query(Document) \
                .filter_by(text=question, answer=answer).count():
            logger.info('Already here: {0} -> {1}'.format(question, answer))
            return
        logger.info('add document: {0} -> {1}'.format(question, answer))

        grams = self._get_grams(session, question, make=True)

        doc = Document(question, answer)
        doc.grams = list(grams)
        self._recalc_idfs(session, grams)

        session.add(doc)
        session.commit()

    def get_best_answer(self, query):
        """Get best answer to a question.

        :param query: A question to get an answer
        :type query: :class:`str`

        :returns: An answer to a question
        :rtype: :class:`str`

        :raises: :class:`NoAnswerError` when can not found answer to a question

        """
        query = to_unicode(query)
        session = self.Session()

        grams = self._get_grams(session, query)
        if not grams:
            raise NoAnswerError('Can not found answer')

        documents = set([doc for gram in grams for doc in gram.documents])

        self._recalc_idfs(session, grams)

        idfs = dict((gram.gram, gram.idf) for gram in grams)

        docs = dict(
            (doc.answer, _cosine_measure(idfs, self._get_tf_idfs(doc)))
            for doc in documents)
        docs = dict((key, val) for (key, val) in docs.items() if val)

        session.commit()

        try:
            max_ratio = max(docs.values())
            answers = [answer for answer in docs.keys()
                       if docs.get(answer) == max_ratio]

            answer = random.choice(answers)
            logger.debug('{0} -> {1} ({2})'.format(query, answer, max_ratio))
            return (answer, max_ratio)
        except ValueError:
            raise NoAnswerError('Can not found answer')
        finally:
            session.commit()

    def recreate_grams(self):
        """Re-create grams for database.

        In normal situations, you never need to call this method.
        But after migrate DB, this method is useful.

        :param session: DB session
        :type session: :class:`sqlalchemt.orm.Session`

        """

        session = self.Session()

        for document in session.query(Document).all():
            logger.info(document.text)
            grams = self._get_grams(session, document.text, make=True)
            document.grams = list(grams)

        broken_links = session.query(Gram) \
            .filter(~Gram.documents.any()).all()
        for gram in broken_links:
            session.delete(gram)

        session.commit()

    def recalc_idfs(self, grams=None):
        session = self.Session()
        self._recalc_idfs(session, grams)
        session.commit()

    def _recalc_idfs(self, session, grams=None):
        """Re-calculate idfs for database.

        calculating idfs for gram is taking long time.
        So I made it calculates idfs for some grams.
        If you want make accuracy higher, use this with grams=None.

        :param session: DB session
        :type session: :class:`sqlalchemt.orm.Session`

        :param grams: grams that you want to re-calculating idfs
        :type grams: A set of :class:`Gram`

        """

        if not grams:
            grams = session.query(Gram)
        for gram in grams:
            orig_idf = gram.idf
            gram.idf = self._get_idf(session, gram)
            logger.debug('Recalculating {} {} -> {}'.format(
                gram.gram, orig_idf, gram.idf))

    def get_count(self):
        """Get count of :class:`Document`.

        :param session: DB session
        :type session: :class:`sqlalchemt.orm.Session`

        """
        return self.Session.query(Document).count()

    def _get_grams(self, session, text, make=False):
        grams = set()

        if len(text) < GRAM_LENGTH:
            gram_obj = session.query(Gram).filter_by(gram=text).first()
            if gram_obj:
                grams.add(gram_obj)
            elif make:
                gram_obj = Gram(text)
                session.add(gram_obj)
                grams.add(gram_obj)
        else:
            for i in range(len(text) - GRAM_LENGTH + 1):
                gram = text[i:i+GRAM_LENGTH]
                gram_obj = session.query(Gram).filter_by(gram=gram).first()
                if gram_obj:
                    grams.add(gram_obj)
                else:
                    gram_obj = Gram(gram)
                    grams.add(gram_obj)
                    if make:
                        session.add(gram_obj)

        return grams

    @staticmethod
    def _get_tf(gram, document):
        if isinstance(gram, Gram):
            gram = gram.gram

        gram = make_string(gram)

        return document.text.count(gram) + document.answer.count(gram)

    def _get_idf(self, session, gram):
        all_count = session.query(Document).count()
        d_count = len(gram.documents)
        return math.log((all_count / (1.0 + d_count)) + 1)

    def _get_tf_idfs(self, document):
        tf_idfs = dict(
            (gram.gram, self._get_tf(gram, document) * gram.idf)
            for gram in document.grams
            if gram.idf is not None)
        return tf_idfs


def _cosine_measure(v1, v2):
    intersection = set(v1.keys()) & set(v2.keys())
    numerator = sum([v1[x] * v2[x] for x in intersection])

    sum1 = sum([v1[x]**2 for x in v1.keys() if v1[x] is not None])
    sum2 = sum([v2[y]**2 for y in v2.keys() if v2[y] is not None])

    denominator = math.sqrt(sum1) * math.sqrt(sum2)

    try:
        return numerator / denominator
    except ZeroDivisionError:
        return 0


class NoAnswerError(Exception):
    """Raises when autotweet can not found best answer to a question.

    :param msg: A message for the exception
    :type msg: :class:`str`

    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)



from __future__ import unicode_literals

import logging

root_logger = logging.getLogger('autotweet')

logging.basicConfig(
    format='%(asctime)s {%(module)s:%(levelname)s}: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S')


def set_level(level):
    root_logger.setLevel(level)


get_logger = root_logger.getChild



# coding: utf-8
""":mod:`autotweet.command` --- CLI interface
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides abillities to connect to telegram.

"""
from __future__ import unicode_literals

import re

from telegram.ext import (
    BaseFilter, CommandHandler, Filters, MessageHandler, Updater)

from .learning import DataCollection, NoAnswerError
from .logger_factory import get_logger
from .twitter import strip_tweet


class ReplyFilter(BaseFilter):
    def filter(self, message):
        return bool(message.reply_to_message)


Filters.reply = ReplyFilter()

logger = get_logger(__name__)


class TelegramBot(object):
    def __init__(self, db_uri, token, threshold, learning=True, answering=True):
        self._make_updater(token)

        self.threshold = threshold
        self.data_collection = DataCollection(db_uri)

        self._init_handlers()
        if learning:
            self.enable_learning()
        if answering:
            self.enable_answering()

    def run(self):
        logger.info('Starting with {} documents.'.format(
            self.data_collection.get_count()))
        self.me = self.updater.bot.get_me()
        self.updater.start_polling()
        self.updater.idle()

    def learning_handler(self, bot, update):
        question = strip_tweet(update.message.reply_to_message.text)
        answer = strip_tweet(update.message.text, remove_url=False)
        self.data_collection.add_document(question, answer)

    def answering_handler(self, bot, update):
        question = strip_tweet(update.message.text)
        try:
            answer, ratio = self.data_collection.get_best_answer(question)
            if (ratio > self.threshold or
                    self._is_necessary_to_reply(bot, update)):
                logger.info('{} -> {}'.format(question, answer))
                update.message.reply_text(answer)
        except NoAnswerError:
            logger.debug('No answer to {}'.format(update.message.text))
            if self._is_necessary_to_reply(bot, update):
                update.message.reply_text(r'¯\_(ツ)_/¯')

    def leave_handler(self, bot, update):
        logger.info('Leave from chat {}'.format(update.message.chat_id))
        bot.leave_chat(update.message.chat_id)

    def enable_learning(self):
        logger.debug('Enabling learning handler.')
        self.dispatcher.add_handler(
            MessageHandler(Filters.reply, self.learning_handler))

    def enable_answering(self):
        logger.debug('Enabling answer handler.')
        self.dispatcher.add_handler(
            MessageHandler(Filters.text, self.answering_handler))

    def _make_updater(self, token):
        self.updater = Updater(token)
        self.dispatcher = self.updater.dispatcher

    def _init_handlers(self):
        self.dispatcher.add_handler(CommandHandler('leave', self.leave_handler))

    def _is_necessary_to_reply(self, bot, update):
        message = update.message

        if message.chat.type == 'private':
            logger.debug('{} type private'.format(message.text))
            return True

        matched = re.search(r'@{}\b'.format(self.me.username), message.text)
        result = bool(matched)
        if result:
            logger.debug('{} mentioned me.'.format(message.text))
            return True

        return False


def start_bot(token, db_uri, threshold, learning=True, answering=True):
    bot = TelegramBot(db_uri, token, threshold, learning, answering)
    bot.run()
